#!/usr/bin/env node

process.chdir(__dirname)
process.chdir('..')

var fs = require('fs')

try {
    var uglifyJs = require('uglify-js')
} catch (e) {
    console.log('ERROR: module uglify-js not found. run "npm install uglify-js" to install.')
    process.exit(1)
}

var files = [
    'AbsoluteDiv.js',
    'ApplyModifier.js',
    'ArrayCall.js',
    'ArrowDownHint.js',
    'ArrowDownHintToolButton.js',
    'ArrowUpHint.js',
    'ArrowUpHintToolButton.js',
    'AwakeRemoteAPI.js',
    'BottomToolBar.js',
    'ButtonBar.js',
    'Button.js',
    'CheckBox.js',
    'Cookie.js',
    'DeleteFilesConfirmDialog.js',
    'Description.js',
    'Dialog.js',
    'Div.js',
    'DocumentStatisticsDialog.js',
    'ExpandIcon.js',
    'ExportSessionDialog.js',
    'FileField.js',
    'Filename.js',
    'FileNameField.js',
    'Hint.js',
    'HintToolButton.js',
    'Icon.js',
    'ID.js',
    'ImportSessionDialog.js',
    'KeyCodes.js',
    'Languages.js',
    'LeftLabel.js',
    'LeftLabelTextField.js',
    'LicenseDialog.js',
    'LoadingIcon.js',
    'LoadingText.js',
    'MessagePane.js',
    'NewFolderDialog.js',
    'NewNetworkFolderDialog.js',
    'NewZviniFolderDialog.js',
    'Notification.js',
    'NotificationBar.js',
    'OpenFileDialog.js',
    'Path.js',
    'Preferences.js',
    'QueryString.js',
    'RemoteAPI.js',
    'RenameDialog.js',
    'ReplaceFileConfirmDialog.js',
    'ResetSessionConfirmDialog.js',
    'RevertFileConfirmDialog.js',
    'RootPane.js',
    'SaveChangesConfirmDialog.js',
    'SaveFileDialog.js',
    'SearchFilesDialog.js',
    'ShareSessionDialog.js',
    'SidePane.js',
    'Spinner.js',
    'StringFormat.js',
    'TextField.js',
    'TextNode.js',
    'Throttle.js',
    'ToggleToolButton.js',
    'ToolBar.js',
    'ToolBarSeparator.js',
    'ToolButton.js',
    'TopLabel.js',
    'TopLabelFileField.js',
    'TopLabelTextField.js',
    'UrlencodedXHR.js',
    'UTF8.js',
    'XHR.js',
    'AboutDialog/Dialog.js',
    'AboutDialog/Label.js',
    'AboutDialog/VersionLabel.js',
    'File/Bar.js',
    'File/ErrorBar.js',
    'File/File.js',
    'File/GoToLineBar.js',
    'File/LineNumbers.js',
    'File/OverwriteBar.js',
    'File/PreviewPane.js',
    'File/ReplaceBar.js',
    'File/SearchBar.js',
    'File/StatusBar.js',
    'FileList/EmptyFolder.js',
    'FileList/FileItem.js',
    'FileList/FolderItem.js',
    'FileList/List.js',
    'FileTabs/ScrollBar.js',
    'FileTabs/ScrollBarArrow.js',
    'FileTabs/Tab.js',
    'FileTabs/Tabs.js',
    'Menu/CheckItem.js',
    'Menu/Group.js',
    'Menu/Item.js',
    'MenuBar/Bar.js',
    'MenuBar/Item.js',
    'Modifiers/EncodeBase64.js',
    'Modifiers/ChangeCase.js',
    'Modifiers/EncodeHex.js',
    'PreferencesDialog/Dialog.js',
    'PreferencesDialog/GeneralTab.js',
    'PreferencesDialog/LanguageTab.js',
    'RichTextarea/AltDownModule.js',
    'RichTextarea/AltUpModule.js',
    'RichTextarea/CtrlBackspaceModule.js',
    'RichTextarea/CtrlDeleteModule.js',
    'RichTextarea/CtrlDModule.js',
    'RichTextarea/CtrlDownModule.js',
    'RichTextarea/CtrlEndModule.js',
    'RichTextarea/CtrlHomeModule.js',
    'RichTextarea/CtrlLeftModule.js',
    'RichTextarea/CtrlLeftSquareBracketModule.js',
    'RichTextarea/CtrlRightModule.js',
    'RichTextarea/CtrlRightSquareBracketModule.js',
    'RichTextarea/CtrlUpModule.js',
    'RichTextarea/DownModule.js',
    'RichTextarea/EndModule.js',
    'RichTextarea/HomeModule.js',
    'RichTextarea/LeftModule.js',
    'RichTextarea/EnterModule.js',
    'RichTextarea/RightModule.js',
    'RichTextarea/ShiftCtrlDownModule.js',
    'RichTextarea/ShiftCtrlLeftModule.js',
    'RichTextarea/ShiftCtrlLeftSquareBracketModule.js',
    'RichTextarea/ShiftCtrlRightModule.js',
    'RichTextarea/ShiftCtrlRightSquareBracketModule.js',
    'RichTextarea/ShiftCtrlUpModule.js',
    'RichTextarea/ShiftDownModule.js',
    'RichTextarea/ShiftEndModule.js',
    'RichTextarea/ShiftHomeModule.js',
    'RichTextarea/ShiftLeftModule.js',
    'RichTextarea/ShiftRightModule.js',
    'RichTextarea/ShiftTabModule.js',
    'RichTextarea/ShiftUpModule.js',
    'RichTextarea/TabModule.js',
    'RichTextarea/Textarea.js',
    'RichTextarea/UpModule.js',
    'String/FindCtrlDownIndex.js',
    'String/FindCtrlLeftIndex.js',
    'String/FindCtrlLeftSquareBracketIndex.js',
    'String/FindCtrlRightIndex.js',
    'String/FindCtrlRightSquareBracketIndex.js',
    'String/FindCtrlUpIndex.js',
    'String/FindDownIndex.js',
    'String/FindLineEnd.js',
    'String/FindLineStart.js',
    'String/FindUpIndex.js',
    'Tabs/Tabs.js',
    'Tabs/Tab.js',
    'Languages/de/Terms.js',
    'Languages/en/Terms.js',
    'Languages/ka/Terms.js',
    'Main.js',
]

var source = '(function () {\n'
files.forEach(function (file) {
    source += fs.readFileSync('javascript/' + file, 'utf8') + ';\n'
})
source += '\n})()'

fs.writeFileSync('combined.js', source)

var compressedSource = uglifyJs.minify({
    'combined.js': source,
}, {
    toplevel: true,
    warnings: true,
    output: { max_line_len: 1024 },
}).code

fs.writeFileSync('compressed.js', compressedSource)
